"""
Module with MagmaFrame classes and subclasses.
"""

from MagmaPandas.MagmaFrames.clinopyroxene import *
from MagmaPandas.MagmaFrames.ilmenite import *
from MagmaPandas.MagmaFrames.magmaFrame import *
from MagmaPandas.MagmaFrames.magnetite import *
from MagmaPandas.MagmaFrames.melt import *
from MagmaPandas.MagmaFrames.olivine import *
from MagmaPandas.MagmaFrames.plagioclase import *
